"""
Utilities for downloading files.
"""
import os
from pathlib import Path
from sys import stderr

import requests
from tqdm import tqdm


def download(url: str, output: Path) -> Path:
    """
    Download a file from a given URL.

    .. note:: Requires the :ref:`download extra <index:download>`.
    """
    print(f'Downloading "{url}"...', file=stderr)
    data_stream = requests.get(url, stream=True, timeout=30)
    total_size = int(data_stream.headers.get('content-length', 0))

    with tqdm(
        total=total_size, leave=False,
        unit='iB', unit_scale=True, unit_divisor=1024,
        disable='GITHUB_ACTIONS' in os.environ,
    ) as progress_bar:
        with open(output, 'wb') as output_file:
            for data in data_stream.iter_content(chunk_size=1024):
                progress_bar.update(output_file.write(data))
    return output



"""
Utilities for testing.
"""
from collections.abc import Callable
from operator import methodcaller
from typing import Any


def hide_traceback[Function: Callable[..., Any]](
    function: Function,
    error: type[Exception] = AssertionError,
) -> Function:
    """
    Decorator for hiding pytest failure tracebacks.

    See the :ref:`pytest documentation <pytest:__tracebackhide__>` for more information.
    """
    function.__globals__['__tracebackhide__'] = methodcaller('errisinstance', error)
    return function



from collections.abc import Iterable


def unique[T](iterable: Iterable[T]) -> Iterable[T]:
    """
    Yield unique elements from an iterable.
    """
    seen = set()
    for element in iterable:
        if element not in seen:
            seen.add(element)
            yield element



import logging
import os
import shutil
import stat
from pathlib import Path
from tempfile import mkdtemp
from zipfile import ZipFile

logger = logging.getLogger(__name__)


def tmp_path(prefix: str, suffix: str) -> Path:
    """
    Create a temporary path.
    """
    path = Path(mkdtemp(prefix=f'{prefix}_', suffix=f'_{suffix}'))
    path.mkdir(parents=True, exist_ok=True)  # ensure the path exists
    logger.debug(f'Using temporary path "{path}/"')
    return path


def unzip(archive: Path) -> Path:
    """
    Extract an archive.
    """
    with ZipFile(archive) as archive_file:
        for member_info in archive_file.infolist():
            extracted_path = archive_file.extract(member_info, archive.parent / archive.stem)

            attributes = member_info.external_attr >> 16
            attributes &= stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO  # extract permission bits
            if member_info.is_dir():
                os.chmod(extracted_path, mode=0o755)
            elif attributes:
                os.chmod(extracted_path, mode=attributes)
    return archive


def move(source: Path, destination: Path) -> Path:
    """
    Move a file from source to destination.
    """
    destination.parent.mkdir(parents=True, exist_ok=True)  # create parent folders
    shutil.rmtree(destination, ignore_errors=True)  # remove existing content
    shutil.move(str(source), str(destination))
    return destination



"""
Utilities for viewing project configuration.
"""
import tomllib
from decimal import Decimal
from pathlib import Path
from typing import Any

PATH = Path('pyproject.toml')
#: The ``pyproject.toml`` configuration of the current project.
PYPROJECT: dict[str, Any] = (
    tomllib.loads(PATH.read_text(encoding='utf-8'), parse_float=Decimal)
    if PATH.exists() else {}
)


def tool_config(name: str) -> Any:
    """
    Return the ``pyproject.toml`` configuration of a given tool.
    """
    return PYPROJECT.get('tool', {}).get(name, {})



"""
Utilities for managing imports.
"""
from importlib import import_module
from types import ModuleType


def try_import(name: str) -> ModuleType | None:
    """
    Return the imported module, or :data:`None` if it was not found.
    """
    try:
        return import_module(name=name)
    except ImportError:
        return None
